@@ -38,9 +38,13 @@ struct _EmpathyAccountWidgetUIDetails {
 };
 
 
-void       empathy_account_widget_handle_params     (EmpathyAccountWidget *self,
-						     const gchar *first_widget,
-						     ...);
+void empathy_account_widget_handle_params (EmpathyAccountWidget *self,
+    const gchar *first_widget,
+    ...);
+
+void empathy_account_widget_setup_widget (EmpathyAccountWidget *self,
+    GtkWidget *widget,
+    const gchar *param_name);
 
 G_END_DECLS
 
