@@ -61,11 +61,6 @@
 #define DEBUG_FLAG EMPATHY_DEBUG_CHAT
 #include <libempathy/empathy-debug.h>
 
-typedef struct {
-	EmpathyChatWindow *window;
-	EmpathyChat *chat;
-} NotificationData;
-
 #define GET_PRIV(obj) EMPATHY_GET_PRIV (obj, EmpathyChatWindow)
 typedef struct {
 	EmpathyChat *current_chat;
@@ -79,7 +74,6 @@ typedef struct {
 	GtkWidget   *dialog;
 	GtkWidget   *notebook;
 	NotifyNotification *notification;
-	NotificationData *notification_data;
 
 	GtkTargetList *contact_targets;
 	GtkTargetList *file_targets;
@@ -333,18 +327,21 @@ chat_window_menu_context_update (EmpathyChatWindowPriv *priv,
 {
 	gboolean first_page;
 	gboolean last_page;
+	gboolean wrap_around;
 	gboolean is_connected;
 	gint     page_num;
 
 	page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (priv->notebook));
 	first_page = (page_num == 0);
 	last_page = (page_num == (num_pages - 1));
+	g_object_get (gtk_settings_get_default (), "gtk-keynav-wrap-around",
+		      &wrap_around, NULL);
 	is_connected = empathy_chat_get_tp_chat (priv->current_chat) != NULL;
 
-	DEBUG ("Update window : Menu Contexts (Tabs & Conv)");
-
-	gtk_action_set_sensitive (priv->menu_tabs_next, TRUE);
-	gtk_action_set_sensitive (priv->menu_tabs_prev, TRUE);
+	gtk_action_set_sensitive (priv->menu_tabs_next, (!last_page ||
+							 wrap_around));
+	gtk_action_set_sensitive (priv->menu_tabs_prev, (!first_page ||
+							 wrap_around));
 	gtk_action_set_sensitive (priv->menu_tabs_detach, num_pages > 1);
 	gtk_action_set_sensitive (priv->menu_tabs_left, !first_page);
 	gtk_action_set_sensitive (priv->menu_tabs_right, !last_page);
@@ -387,8 +384,6 @@ chat_window_contact_menu_update (EmpathyChatWindowPriv *priv,
 		"/chats_menubar/menu_contact");
 	orig_submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (menu));
 
-	DEBUG ("Update window : Contact Menu");
-
 	if (orig_submenu == NULL || !GTK_WIDGET_VISIBLE (orig_submenu)) {
 		submenu = empathy_chat_get_contact_menu (priv->current_chat);
 		gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu), submenu);
@@ -487,8 +482,6 @@ chat_window_title_update (EmpathyChatWindowPriv *priv)
 {
 	gchar *name;
 
-	DEBUG ("Update window : Title");
-
 	name = get_window_title_name (priv);
 	gtk_window_set_title (GTK_WINDOW (priv->dialog), name);
 	g_free (name);
@@ -504,8 +497,6 @@ chat_window_icon_update (EmpathyChatWindowPriv *priv)
 
 	n_chats = g_list_length (priv->chats);
 
-	DEBUG ("Update window : Icon");
-
 	/* Update window icon */
 	if (priv->chats_new_msg) {
 		gtk_window_set_icon_name (GTK_WINDOW (priv->dialog),
@@ -537,8 +528,6 @@ chat_window_close_button_update (EmpathyChatWindowPriv *priv,
 	GtkWidget *chat_close_button;
 	gint       i;
 
-	DEBUG ("Update window : Close Button");
-
 	if (num_pages == 1) {
 		chat = gtk_notebook_get_nth_page (GTK_NOTEBOOK (priv->notebook), 0);
 		chat_close_button = g_object_get_data (G_OBJECT (chat),
@@ -562,8 +551,6 @@ chat_window_update (EmpathyChatWindow *window)
 
 	num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (priv->notebook));
 
-	DEBUG ("Update window");
-
 	/* Update Tab menu */
 	chat_window_menu_context_update (priv,
 					 num_pages);
@@ -1027,14 +1014,18 @@ chat_window_tabs_next_activate_cb (GtkAction         *action,
 	EmpathyChatWindowPriv *priv;
 	EmpathyChat           *chat;
 	gint                  index_, numPages;
+	gboolean              wrap_around;
 
 	priv = GET_PRIV (window);
 
+	g_object_get (gtk_settings_get_default (), "gtk-keynav-wrap-around",
+                      &wrap_around, NULL);
+
 	chat = priv->current_chat;
 	index_ = gtk_notebook_get_current_page (GTK_NOTEBOOK (priv->notebook));
 	numPages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (priv->notebook));
 
-	if (index_ == (numPages - 1)) {
+	if (index_ == (numPages - 1) && wrap_around) {
 		gtk_notebook_set_current_page (GTK_NOTEBOOK (priv->notebook), 0);
 		return;
 	}
@@ -1049,14 +1040,18 @@ chat_window_tabs_previous_activate_cb (GtkAction         *action,
 	EmpathyChatWindowPriv *priv;
 	EmpathyChat           *chat;
 	gint                  index_, numPages;
+	gboolean              wrap_around;
 
 	priv = GET_PRIV (window);
 
+	g_object_get (gtk_settings_get_default (), "gtk-keynav-wrap-around",
+                      &wrap_around, NULL);
+
 	chat = priv->current_chat;
 	index_ = gtk_notebook_get_current_page (GTK_NOTEBOOK (priv->notebook));
 	numPages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (priv->notebook));
 
-	if (index_ <= 0) {
+	if (index_ <= 0 && wrap_around) {
 		gtk_notebook_set_current_page (GTK_NOTEBOOK (priv->notebook), numPages - 1);
 		return;
 	}
@@ -1070,7 +1065,7 @@ chat_window_tabs_left_activate_cb (GtkAction         *action,
 {
 	EmpathyChatWindowPriv *priv;
 	EmpathyChat           *chat;
-	gint                  index_;
+	gint                  index_, num_pages;
 
 	priv = GET_PRIV (window);
 
@@ -1083,6 +1078,9 @@ chat_window_tabs_left_activate_cb (GtkAction         *action,
 	gtk_notebook_reorder_child (GTK_NOTEBOOK (priv->notebook),
 				    GTK_WIDGET (chat),
 				    index_ - 1);
+
+	num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (priv->notebook));
+	chat_window_menu_context_update (priv, num_pages);
 }
 
 static void
@@ -1091,7 +1089,7 @@ chat_window_tabs_right_activate_cb (GtkAction         *action,
 {
 	EmpathyChatWindowPriv *priv;
 	EmpathyChat           *chat;
-	gint                  index_;
+	gint                  index_, num_pages;
 
 	priv = GET_PRIV (window);
 
@@ -1101,6 +1099,9 @@ chat_window_tabs_right_activate_cb (GtkAction         *action,
 	gtk_notebook_reorder_child (GTK_NOTEBOOK (priv->notebook),
 				    GTK_WIDGET (chat),
 				    index_ + 1);
+
+	num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (priv->notebook));
+	chat_window_menu_context_update (priv, num_pages);
 }
 
 static void
@@ -1184,35 +1185,17 @@ chat_window_set_urgency_hint (EmpathyChatWindow *window,
 
 	priv = GET_PRIV (window);
 
-	DEBUG ("Turning %s urgency hint", urgent ? "on" : "off");
 	gtk_window_set_urgency_hint (GTK_WINDOW (priv->dialog), urgent);
 }
 
-static void
-free_notification_data (NotificationData *data)
-{
-	g_object_unref (data->chat);
-	g_slice_free (NotificationData, data);
-}
-
 static void
 chat_window_notification_closed_cb (NotifyNotification *notify,
-				    NotificationData *cb_data)
+				    EmpathyChatWindow *self)
 {
-	EmpathyNotificationClosedReason reason = 0;
-	EmpathyChatWindowPriv *priv = GET_PRIV (cb_data->window);
-
-#ifdef notify_notification_get_closed_reason
-	reason = notify_notification_get_closed_reason (notify);
-#endif
-	if (reason == EMPATHY_NOTIFICATION_CLOSED_DISMISSED) {
-		empathy_chat_window_present_chat (cb_data->chat);
-	}
+	EmpathyChatWindowPriv *priv = GET_PRIV (self);
 
 	g_object_unref (notify);
 	priv->notification = NULL;
-	free_notification_data (cb_data);
-	priv->notification_data = NULL;
 }
 
 static void
@@ -1247,17 +1230,11 @@ chat_window_show_or_update_notification (EmpathyChatWindow *window,
 		notify_notification_update (priv->notification,
 					    header, escaped, NULL);
 	} else {
-		NotificationData *cb_data = cb_data = g_slice_new0 (NotificationData);
-
-		cb_data->chat = g_object_ref (chat);
-		cb_data->window = window;
-
-		priv->notification_data = cb_data;
 		priv->notification = notify_notification_new (header, escaped, NULL, NULL);
 		notify_notification_set_timeout (priv->notification, NOTIFY_EXPIRES_DEFAULT);
 
 		g_signal_connect (priv->notification, "closed",
-				  G_CALLBACK (chat_window_notification_closed_cb), cb_data);
+				  G_CALLBACK (chat_window_notification_closed_cb), window);
 	}
 
 	pixbuf = empathy_notify_manager_get_pixbuf_for_notification (priv->notify_mgr,
@@ -1522,8 +1499,6 @@ chat_window_focus_in_event_cb (GtkWidget        *widget,
 {
 	EmpathyChatWindowPriv *priv;
 
-	DEBUG ("Focus in event, updating title");
-
 	priv = GET_PRIV (window);
 
 	priv->chats_new_msg = g_list_remove (priv->chats_new_msg, priv->current_chat);
@@ -1764,11 +1739,6 @@ chat_window_finalize (GObject *object)
 		notify_notification_close (priv->notification, NULL);
 		g_object_unref (priv->notification);
 		priv->notification = NULL;
-		if (priv->notification_data != NULL)
-			{
-				free_notification_data (priv->notification_data);
-				priv->notification_data = NULL;
-			}
 	}
 
 	if (priv->contact_targets) {
